/* 
 * Simple IPv4 over IPv6 tunnel device.
 * 
 * Written 2000 by Andi Kleen, SuSE Labs. 
 */
#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/netdevice.h>
#include <linux/if_arp.h>
#include <linux/in.h>
#include <linux/in6.h>
#include <linux/icmp.h>
#include <linux/ipv6.h>
#include <linux/icmpv6.h>
#include <linux/skbuff.h>
#include <linux/malloc.h>
#include <net/protocol.h>
#include <net/ip.h>
#include <net/ipv6.h>
#include <net/ip6_route.h>
#include <net/icmp.h>
#include <asm/uaccess.h>
#include <asm/errno.h>

#include "fourtun.h"

#define IPV6_TLV_ENCAP_LIMIT 4

struct tun_limit_hdr { 
	__u8 type;	 
	__u8 len; 	
	__u8 limit;
} __attribute__((packed)); 

struct fourtun_hdr { 
	struct ipv6hdr ip6;
	struct ipv6_destopt_hdr destopt; 
	struct tun_limit_hdr limit; 
	__u8 pad[3]; 
} __attribute__((packed)); 
#define FOURTUN_EXTHDR_LEN (sizeof(struct fourtun_hdr) - sizeof(struct ipv6hdr))

struct fourtun {
	struct fourtun *next; 
	struct fourtun_hdr hdr; 
	int recursion; 
	struct sock *sk; 
	struct device *dev; 
	struct socket *sock; 	/* Socket for fragmentation */ 
	struct net_device_stats stats; 
}; 
#define v6hdr hdr.ip6

static struct in6_addr nulladdr; 
#define is_null_addr(adr) (!memcmp(&nulladdr, &(adr), sizeof(struct in6_addr)))

static struct fourtun_stat fstat; 

/* Hashtable by foreign tunnel endpoint address. */ 
#define HASHTAB_SIZE 16

// XXX: check the quality of the hash function
#define HASH(addr) (((addr)->s6_addr32[0] ^ (addr)->s6_addr32[1] ^ \
		     (addr)->s6_addr32[2] ^ (addr)->s6_addr32[3]) & (HASHTAB_SIZE-1))
static struct fourtun *tunnel_hash[HASHTAB_SIZE];

static void fourtun_hash(struct fourtun *tun) 
{ 
	int hash = HASH(&tun->v6hdr.daddr); 

	start_bh_atomic(); 
	tun->next = tunnel_hash[hash];
	tunnel_hash[hash] = tun;
	end_bh_atomic();
} 

static void fourtun_unhash(struct fourtun *tun) 
{ 
	int hash = HASH(&tun->v6hdr.daddr); 
	struct fourtun *t, **tp; 

	start_bh_atomic(); 
	for (tp = &tunnel_hash[hash], t = *tp; t; tp = &t->next, t = *tp) { 
		if (t == tun) { 
			*tp = t->next; 
			t->next = NULL; 
			end_bh_atomic();
			return;
		} 
	} 
	end_bh_atomic(); 
} 

static inline struct fourtun *fourtun_hash_lookup(struct ipv6hdr *hdr)
{ 
	int hash = HASH(&hdr->saddr); 
	struct fourtun *tun; 

	for (tun = tunnel_hash[hash]; tun; tun = tun->next) { 
		if (!ipv6_addr_cmp(&hdr->saddr, &tun->v6hdr.daddr))
			break; 
	}  
	return tun; 
} 

/* 
 * This should be moved up to the generic Linux TLV parser.
 * This parser is very limited and does not handle HOP-by-HOP or AH
 * The normal stack has already done a parse (or the packet has been locally
 * originated), so don't try too hard to handle border cases.
 */ 
static int __fourtun_check_limit(struct ipv6hdr *hdr, unsigned char *pend) 
{ 
	unsigned char *p, *end, l; 
	struct ipv6_opt_hdr *dst = (struct ipv6_opt_hdr *)(hdr+1);
	p = (unsigned char *) (dst+1); 
	if (p > pend || (end = ((unsigned char *)dst)+ipv6_optlen(dst)) > pend) {
		fstat.bad_pkt++; 
		return -1;
	} 
	for (; p < end; ) { 
		switch (*p) { 
		case IPV6_TLV_PAD0:
			++p; 
			break;
		default: 
		case IPV6_TLV_PADN:
			l = *++p;
			p+=l+1;
			break; 
		case IPV6_TLV_ENCAP_LIMIT:
			if (p[2] <= 1) 
				return -1; 
			p[2]--; 
			return 0; 
		}
	} 
	return 0;
} 

static inline int fourtun_check_limit(struct ipv6hdr *hdr, unsigned char *pend) 
{ 
	if (hdr->nexthdr == NEXTHDR_DEST) 
		return __fourtun_check_limit(hdr,pend);
	return 0;
}

/* 
 * Receive a packet from IPv6 (tunnel exit point) 
 */ 
static int fourtun_rcv(struct sk_buff *skb, unsigned long len) 
{
	struct fourtun *tun; 

	tun = fourtun_hash_lookup(skb->nh.ipv6h); 
	if (tun == NULL) {
		fstat.unmatched_packets++;
		return -EIO;
	}

	if (fourtun_check_limit(skb->nh.ipv6h, skb->end) < 0) { 
		fstat.tunnel_loop++; 
		return -ELOOP; 
	} 

	dst_release(skb->dst); 
	skb->dst = NULL; 
	skb->protocol = ntohs(ETH_P_IP); 
	skb->ip_summed = 0; 
	skb->dev = tun->dev;
	memset(skb->cb, 0, sizeof(skb->cb)); 
	tun->stats.rx_bytes += skb->len; 
	skb_pull(skb, skb->h.raw - skb->data); 
	tun->stats.rx_packets++; 
	netif_rx(skb);
	return 0; 
} 

/* 
 * Receive a ICMPv6 from the tunnel.
 */ 
static void fourtun_err(struct sk_buff *skb, struct ipv6hdr *hdr, 
			 struct inet6_skb_parm *opt,
			 int type, int code, unsigned char *pb, __u32 info) 
{
	struct iphdr *iph; 
	struct fourtun *tun; 

	skb_pull(skb, pb - ((unsigned char *)hdr)); 
	tun = fourtun_hash_lookup(hdr); 
	if (!tun) {
		fstat.unmatched_icmp++; 
		return; 
	} 

	if (skb->len < sizeof(struct iphdr)) {
		fstat.bad_icmp++; 
		return; 
	} 
	
	iph = (struct iphdr *) skb->data; 
	
	switch (type) { 
	case ICMPV6_DEST_UNREACH: 
	case ICMPV6_PARAMPROB: 
	case ICMPV6_TIME_EXCEED:
		type = ICMP_DEST_UNREACH;
		code = ICMP_HOST_UNREACH;
		info = 0; 
		break; 
	case ICMPV6_PKT_TOOBIG:
		fstat.pmtu_rcv++;
		type = ICMP_DEST_UNREACH; 
		code = ICMP_FRAG_NEEDED;
		{ 
			int mtu = htons(((struct icmp6hdr *)iph - 1)->icmp6_mtu); 
			mtu -= sizeof(struct fourtun_hdr); 
			if (mtu < tun->dev->mtu && mtu > 68) { 
				tun->dev->mtu = mtu; 
				fstat.pmtu_rmt_sent++;
				fstat.mtu_change++;
			} else { 
				fstat.pmtu_bad++; 
				return; 
			} 
		} 
		break; 
	default:
		return;
	}

#if 1
	{
		static char zero[sizeof(skb->cb)]; 
		if (!memcmp(skb->cb, zero, sizeof(skb->cb))) { 
			printk(KERN_DEBUG "fourtun: skb->cb not zeroed in icmp\n");
			return;
		}
			
	}
#endif

	fstat.sent_icmp++;

	skb->nh.iph = iph; 
	icmp_send(skb, type, code, info); 
} 

/*
 * Send fragments over a socket.
 * Tunnels violate the ``fragmentation only on the endnodes'' rule,
 * thus the socket is required.
 */ 

static int fourtun_getfrag(const void *data, struct in6_addr *addr,
			 char *buff, unsigned int offset, unsigned int len)
{
	memcpy(buff, data + offset, len); 
	return 0; 
}
				   
static void fourtun_frag_xmit(struct fourtun *tun, struct sk_buff *skb,
			       int pmtu, struct flowi *fl)
{
	if (skb->nh.iph->frag_off & htons(IP_DF)) {
		fstat.pmtu_local_sent++;
		icmp_send(skb, ICMP_DEST_UNREACH, ICMP_FRAG_NEEDED, htons(pmtu)); 
	} else { 
		tun->stats.tx_packets++; 
		tun->stats.tx_bytes += skb->len; 
		ip6_build_xmit(tun->sk, fourtun_getfrag, skb->data, 
			       fl, skb->len, 
			       NULL, 64, MSG_DONTWAIT);
	}
}

/* 
 * Send a tunnel packet (tunnel entry point) 
 */ 
static int fourtun_dev_xmit(struct sk_buff *skb, struct device *dev)
{
	struct dst_entry *dst;
	struct flowi fl; 
	int max_headroom; 
	int oldlen = skb->len; 
	struct fourtun *tunnel = (struct fourtun *) dev->priv; 

	if (!tunnel || skb->protocol != htons(ETH_P_IP)) { 
		++tunnel->stats.tx_errors;
		dev_kfree_skb(skb); 
		return -EIO; 
	}

	if (tunnel->recursion) { 
		++tunnel->stats.collisions; 
		dev_kfree_skb(skb); 
		return -ELOOP;
	}

	/* Paranoia. Handle v6-in-SIT-over-v6 loops. */ 
	if (skb->nh.iph->protocol == IPPROTO_IPV6) {
		fstat.v6_in_v4++;
		if (fourtun_check_limit((struct ipv6hdr *) 
								((char *)skb->nh.iph + (skb->nh.iph->ihl<<2)),
								skb->end) < 0) { 
			fstat.tunnel_loop++;
			dev_kfree_skb(skb); 
			return -ELOOP; 
		}
	} 

	memset(&fl,0,sizeof(fl)); 
	fl.proto = IPPROTO_IPV4; 
	fl.fl6_dst = &tunnel->v6hdr.daddr; 
	fl.fl6_src = &tunnel->v6hdr.saddr;

	// XXX: cache the route

	dst = ip6_route_output(NULL, &fl); 
	if (dst->error) { 
		int err;
		tunnel->stats.tx_carrier_errors++; 
		dev_kfree_skb(skb); 
		err = dst->error;
		dst_release(dst); 
		return err;
	}  

	if (dst->pmtu > tunnel->dev->mtu) {
		fstat.mtu_change++; 
		tunnel->dev->mtu = dst->pmtu;
	}
	
	{
		int mtu = dst->pmtu - sizeof(struct fourtun_hdr); 

		if (skb->len > mtu) { 
			fstat.frags_needed++; 
			fourtun_frag_xmit(tunnel, skb, mtu, &fl); 
			dst_release(dst); 
			dev_kfree_skb(skb);
			return 0;
		} 
	}

	max_headroom = ((dst->dev->hard_header_len+15)&~15) + 
							sizeof(struct fourtun_hdr); 

	if (skb_headroom(skb) < max_headroom || skb_cloned(skb) || skb_shared(skb)) { 
		struct sk_buff *nskb; 
		nskb = skb_realloc_headroom(skb, max_headroom); 
		if (!nskb) { 
			fstat.oom++; 
			dst_release(dst); 
			dev_kfree_skb(skb); 
			return -ENOMEM; 
		}
		if (skb->sk) 
			skb_set_owner_w(nskb, skb->sk); 
		dev_kfree_skb(skb);
		skb = nskb; 
	}

	tunnel->stats.tx_packets++; 
	tunnel->stats.tx_bytes += skb->len; 

	dst_release(skb->dst);
	skb->dst = dst;
	memset(skb->cb, 0, sizeof(skb->cb)); 

	// XXX: could copy ipv4 tos/diffserv token over from the v4 packet.

	{ 
		char *hdr = skb_push(skb, sizeof(struct fourtun_hdr));  
		skb->nh.raw = hdr; 
		memcpy(hdr, &tunnel->hdr, sizeof(struct fourtun_hdr)); 
		skb->nh.ipv6h->payload_len = htons(oldlen + FOURTUN_EXTHDR_LEN); 
	} 

	tunnel->recursion++; 
	ip6_output(skb); 
	tunnel->recursion--; 

	return 0;
}

/* Prebuild the IPv6 header. Only length is filled in later. */ 
static void fourtun_build_hdr(struct fourtun_hdr *fhdr, struct fourtun_args *args)
{ 
	struct ipv6hdr *hdr = &fhdr->ip6;
	
	/* Sanity check */ 
	if (sizeof(struct fourtun_hdr) % 8) { 
		extern void __fourtun_hdr_not_aligned(void); 
		__fourtun_hdr_not_aligned(); 
	} 

	hdr->version = 6; 

	hdr->priority = args->priority; 
	hdr->nexthdr = NEXTHDR_DEST;
	hdr->hop_limit = args->hoplimit; 
	
	// XXX: check if saddr is really local
	// XXX: choose suitable local address if saddr is zero
	
	ipv6_addr_copy(&hdr->daddr, &args->daddr); 
	ipv6_addr_copy(&hdr->saddr, &args->saddr); 
	
	fhdr->destopt.nexthdr = IPPROTO_IPV4; 
	fhdr->destopt.hdrlen = 0;  /* ``... not including the first 8 octets'' */
	
	fhdr->limit.type = IPV6_TLV_ENCAP_LIMIT; 
	fhdr->limit.len = 1;
	fhdr->limit.limit = args->encaplimit;
	
	memset(&fhdr->pad, IPV6_TLV_PAD0, sizeof(fhdr->pad)); 
} 

static void fourtun_get_tunnel(struct fourtun *tunnel, struct fourtun_args *args)
{ 
	struct fourtun_hdr *fhdr = &tunnel->hdr; 
	args->priority = fhdr->ip6.priority; 
	args->hoplimit = fhdr->ip6.hop_limit;
	args->encaplimit = fhdr->limit.limit; 
	ipv6_addr_copy(&args->saddr, &fhdr->ip6.saddr);
	ipv6_addr_copy(&args->daddr, &fhdr->ip6.daddr);
} 

static int fourtun_init(struct device *dev, struct fourtun_args *args,
						struct fourtun *tunnel) 
{ 
	int err;

	fourtun_build_hdr(&tunnel->hdr, args); 

	if (!tunnel->sk) { 
		err = sock_create(PF_INET6, SOCK_RAW, IPPROTO_IPV4, &tunnel->sock);
		if (err < 0) { 
			printk(KERN_ERR "fourtun: cannot create fragment socket: %d\n", 
				   err);
			kfree(tunnel); 
			return err;
		}
		
		tunnel->sk = tunnel->sock->sk;
	}

	tunnel->dev = dev; 

	return 0;
} 

static int fourtun_dev_open(struct device *dev)
{ 
	struct fourtun *tun = (struct fourtun *) dev->priv; 
	if (!tun) 
		return -ENODEV;
	if (is_null_addr(tun->hdr.ip6.saddr) || is_null_addr(tun->hdr.ip6.daddr))
		return -ENETDOWN; 
	return 0; 
} 

static void fourtun_free(struct fourtun *tunnel) 
{ 
	fourtun_unhash(tunnel); 
	if (tunnel->sock) 
		sock_release(tunnel->sock);
	kfree(tunnel); 
} 

static int fourtun_dev_ioctl(struct device *dev, struct ifreq *ifr, int cmd)
{
	int err = 0; 
	struct fourtun_args args;
	struct fourtun *tunnel;

	MOD_INC_USE_COUNT; 

	if (cmd == SIOCGTUNNEL6STAT) { 
		err = copy_to_user(ifr->ifr_ifru.ifru_data, &fstat, 
						   sizeof(struct fourtun_stat)); 
		MOD_DEC_USE_COUNT;
		return err ? -EFAULT : 0;
	} 

	if (copy_from_user(&args, ifr->ifr_ifru.ifru_data, sizeof(args))) { 
		MOD_DEC_USE_COUNT; 
		return -EFAULT; 
	} 

	if (args.version != 1) { 
		MOD_DEC_USE_COUNT; 
		return -EADV; 
	}

	switch (cmd) {
	case SIOCADD6TUNNEL:
		err = -EPERM;
		if (!capable(CAP_NET_ADMIN))
			break; 

		err = -EEXIST; 
		if (dev->priv) 
			break; 

		tunnel = kmalloc(sizeof(struct fourtun), GFP_KERNEL);
		err = -ENOMEM; 
		if (!tunnel) 
			break; 
		memset(tunnel,0,sizeof(struct fourtun)); 
		err = fourtun_init(dev, &args, tunnel);
		if (err)
			break; 

		dev->priv = tunnel; 
		fourtun_hash(tunnel); 
		err = 0;
		MOD_INC_USE_COUNT; 
		break; 

	case SIOCDEL6TUNNEL: 
		err = -EPERM;
		if (!capable(CAP_NET_ADMIN))
			break; 

		err = -ENODEV; 
		if (!dev->priv)
			break; 
		
		fourtun_free((struct fourtun *) xchg(&dev->priv, NULL));
		err = 0; 
		MOD_DEC_USE_COUNT; 
		break; 

	case SIOCGET6TUNNEL: 
		err = -ENETDOWN; 
		if (!dev->priv)
			break; 
		fourtun_get_tunnel((struct fourtun *) dev->priv, &args); 
		err = -EFAULT;
		if (copy_to_user(ifr->ifr_ifru.ifru_data, &args, sizeof(args)))
			break; 
		err = 0; 
		break; 

	case SIOCCHG6TUNNEL:
		err = -ENODEV; 
		tunnel = dev->priv; 
		if (!tunnel) 
			break;
		err = fourtun_init(dev, &args, tunnel); 
		if (!err) {
			fourtun_unhash(tunnel);  
			fourtun_hash(tunnel);
		} 
		break; 

	default:
		err = -EIO; 
		break; 
	}
		
	MOD_DEC_USE_COUNT; 
	return err;	
} 

static void fourtun_dev_destroy(struct device *dev)
{
} 

static struct net_device_stats *fourtun_get_stats(struct device *dev) 
{
	struct fourtun *tun = (struct fourtun *) dev->priv; 
	return tun ?  &tun->stats : NULL; 
} 

static int fourtun_dev_init(struct device *dev)
{ 
	dev->priv = NULL; 

	dev->open = fourtun_dev_open; 
	dev->destructor = fourtun_dev_destroy; 
	dev->hard_start_xmit = fourtun_dev_xmit;
	dev->do_ioctl = fourtun_dev_ioctl; 
	dev->get_stats = fourtun_get_stats; 
	dev_init_buffers(dev); 
	dev->type		= ARPHRD_TUNNEL;
	// XXX: compute route and linkdev and use that for mtu/hhl 
	dev->hard_header_len 	= LL_MAX_HEADER + sizeof(struct fourtun_hdr); 
	dev->mtu		= 1500 - sizeof(struct fourtun_hdr);
 	dev->flags		= IFF_NOARP;
	dev->iflink		= 0;
	dev->addr_len		= 4;
	return 0; 
} 

static int numdev = 4;
static struct device **fourtun_devices;  

static struct device fourtun_device_tmpl = { 
	init: fourtun_dev_init
}; 

static struct inet6_protocol fourtun_protocol = { 
	handler: fourtun_rcv, 
	err_handler: fourtun_err, 
	protocol: IPPROTO_IPV4,
	name: "fourtun"
};  

MODULE_AUTHOR("Andi Kleen/SuSE Labs"); 
MODULE_PARM(numdev,"i"); 

int init_module(void)
{
	int c;
	fourtun_devices = kmalloc(sizeof(struct device *) * numdev, GFP_KERNEL); 
	if (!fourtun_devices) 
		return -ENOMEM; 
	for (c = 0; c < numdev; ++c) { 
		struct device *dev = kmalloc(sizeof(struct device) + 20, GFP_KERNEL); 
		if (!dev) 
			break; 
		fourtun_devices[c] = dev; 
		memcpy(dev, &fourtun_device_tmpl, sizeof(struct device)); 
		dev->name = (char *) (dev+1);
		sprintf(dev->name, "fourtun%d", c); 
		register_netdev(dev);  
	} 
	inet6_add_protocol(&fourtun_protocol);
	printk(KERN_INFO "Fourtun 0.1 (%s) for %d devices loaded.\n", 
		   __DATE__, numdev);
	return 0; 
}

void cleanup_module(void)
{
	int c; 
	for (c = 0; c < numdev; c++) { 
		struct fourtun *tun; 
		struct device *dev = fourtun_devices[c]; 
		if (!dev)
			continue; 
		tun = (struct fourtun *)(fourtun_devices[c]->priv);
		if (tun) { 
			dev->priv = NULL; 
			fourtun_free(tun); 
		} 
		unregister_netdev(dev);  
		kfree(dev); 
		fourtun_devices[c] = NULL; 
	} 
	inet6_del_protocol(&fourtun_protocol); 
	printk(KERN_INFO "Fourtun unloaded\n"); 
}
